.sidebar {
    border-width: 1px 0 0 0;
    border-style: solid;
    border-color: shade (@theme_bg_color, 0.66);
    border-radius: 0;
}

.new-item-list {
    background-color: transparent;
}

.pane-separator {
    border-width: 1px 1px 0 0;
    border-style: solid;
    border-color: shade (@theme_bg_color, 0.66);
}

levelbar .strength-weak {
  background-color: #cc0000;
  border-color: #cc0000;
}

levelbar .strength-low {
  background-color: #f5ce00;
  border-color: #f5ce00;
}

levelbar .strength-medium,
levelbar .strength-good,
levelbar .strength-high {
  background-color: #73d216;
  border-color: #73d216;
}
